
# SynPlot - visualization of alignments of syntenic
#           genomic dna sequences using PostScript (TM)
# 
# Copyright (C) 2000 James G.R. Gilbert
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

### SynPlot ###

package SynPlot;

use strict;
use Carp;
use SynPlot::Sequence;
use SynPlot::Summary;
use SynPlot::Feature;
use SynPlot::State;
use vars '@ISA';
@ISA = ('SynPlot::State');

sub new {
    my( $pkg ) = @_;
    
    return bless {}, $pkg;
}

sub name {
    my( $plot, $arg ) = @_;
    
    if ($arg) {
        $plot->{'_name'} = $arg;
    }
    return $plot->{'_name'} || 'Syntenic_Plot';
}

sub landscape {
    my( $plot, $arg ) = @_;
    
    if (defined $arg) {
        $plot->{'_landscape'} = $arg;
    }
    return $plot->{'_landscape'} || 0;
}

sub edge {
    my( $plot, $arg ) = @_;
    
    if (defined $arg) {
        $plot->{'_edge'} = $arg;
    }
    return $plot->{'_edge'} || 36;
}

sub add_element {
    my( $plot, $ele ) = @_;

    $plot->{'_element'} ||= [];
    if ($ele) {
        push @{$plot->{'_element'}}, $ele;
    }
    return scalar @{$plot->{'_element'}};
}

sub each_element {
    my( $plot ) = @_;

    $plot->{'_element'} ||= [];
    return @{$plot->{'_element'}};
}

BEGIN {
    foreach my $type (qw( Sequence Summary Feature )) {
        my $func = "get_$type";
        my $pkg = "SynPlot::$type";
        
        {
            no strict 'refs';
            
            *$func = sub {
                my( $plot, $name ) = @_;
                
                my @are = grep $_->isa($pkg), $plot->each_element;
                if ($name) {
                    @are = grep $_->name eq $name, @are;
                }
                return wantarray ? @are : $are[0];
            }
        }
    }
}

sub create_summary {
    my( $plot ) = @_;
    
    my $summ = 'SynPlot::Summary'->new;
    if ($summ->analyse($plot->get_Sequence)) {
        $plot->add_element($summ);
        return $summ;
    } else {
        confess("Summary creation failed");
    }
}

sub zero_state {
    my( $plot ) = @_;

    my $state = 'SynPlot::State'->new;
    $plot->state($state);
    foreach my $e ($plot->each_element) {
        $e->state($state);
    }
}

sub row_spacing {
    my( $plot, $arg ) = @_;
    
    if ($arg) {
        $plot->{'_row_spacing'} = $arg;
    }
    return $plot->{'_row_spacing'} || 40;
}

sub row_height {
    my( $plot ) = @_;
    
    my $height = 0;
    foreach my $e ($plot->each_element) {
        $height += $e->height;
    }
    return $height;
}

sub draw {
    my( $plot ) = @_;
    
    # Set up A4 page with 36 point (0.5") border
    my( $w, $h ) = (595, 841);
    my( $page_width, $page_height );
    if ($plot->landscape) {
        $page_width  = $h;
        $page_height = $w;
    } else {
        $page_width  = $w;
        $page_height = $h;
    }

    my $edge = $plot->edge;
    my $canvas_width  = $page_width  - ($edge * 2);
    my $canvas_height = $page_height - ($edge * 2);
    $plot->canvas_width ($canvas_width);
    $plot->canvas_height($canvas_height);
    $plot->set_scaling;

    # Header    
    $plot->header($w, $h);

    # Possibly multiple pages
    
    # Multiple rows
    my $origin = $canvas_height + $edge;
    $plot->output("\n$edge $origin translate\n");
    my $row_spacing = $plot->row_spacing;
    
    # next_row sets state parameters so that each element
    # knows which portion of itself to draw.
    while (my $row_num = $plot->next_row) {

        # Print out each of the elements
        ELEMENT: foreach my $e ($plot->each_element) {
            my $y = $e->height;
            next ELEMENT unless $y;     # Ignore non-drawable rows
            $plot->output("\n0 -$y translate\n");
            $e->draw($plot);    # Drawing method may need further elements from plot
        }
        $plot->output("\n% Space between row\n0 -$row_spacing translate\n");
    }
    
    # Footer
    $plot->footer;
}

sub set_scaling {
    my( $plot ) = @_;
    
    my( $nw );
    unless ($nw = $plot->nuc_width) {
        $nw = $plot->guess_a_nuc_width;
    }
    my $scale = $plot->canvas_width / $nw;
    $plot->scale($scale);
    
    if (my $left = $plot->left_offset) {
        $plot->from($left);
    }
}

sub guess_a_nuc_width {
    my( $plot ) = @_;
    
    my $r_h   = $plot->row_height;
    my $r_s   = $plot->row_spacing;
    my $y_max = $plot->canvas_height;
    
    # Maximum number of rows which will fit on canvas
    my $max_rows = 0;
    for (my $y = $r_h; $y < $y_max; $y += $r_s + $r_h) {
        $max_rows++;
    }
    confess("Can't fit any rows on page") if $max_rows == 0;
    
    # Length of the portion of the sequence being drawn,
    # or the whole sequence.
    my $length = $plot->offset_length || $plot->get_Sequence->seq_length;
    
    # Number of residues per row is length divided
    # by max_rows, rounded up to nearest 1000
    my $nuc_width = (1 + int($length / $max_rows / 1000)) * 1000;
        
    $plot->nuc_width($nuc_width);
}

sub next_row {
    my( $plot ) = @_;
    
    # Reset row number and return undef if we're at the end
    my $upto = $plot->upto || 0;
    my $seq_length = $plot->right_offset || $plot->get_Sequence->seq_length;
    if ($upto >= $seq_length) {
        $plot->row_number(0);
        return;
    }
    
    # Otherwise increment the row number
    my $row = $plot->row_number;
    $row++;
    $plot->row_number($row);
    
    # Set the left hand nucleotide of this row
    my $row_length = $plot->nuc_width
        or confess "Standard nucleotide width not set for plot";
    my $from = $plot->from;
    if ($row == 1) {
        $from = $plot->left_offset || 1;
    } else {
        $from += $row_length;
    }
    $plot->from($from);
    
    # Set the right hand nucleotide of this row
    $upto = $from + $row_length - 1;
    if ($upto > $seq_length) {
        $upto = $seq_length;
    }
    $plot->upto($upto);

    $plot->row_length($upto - $from + 1);
    
    return $row;
}

sub header {
    my( $plot, $width, $height ) = @_;
    
    my $name = $plot->name;
    my $date = creation_date();
    
    my $b_b = 29;  # Distance of bounding box from page edge
    my $b_x = $width  - $b_b;
    my $b_y = $height - $b_b;
    
    $plot->output(qq{%!PS-Adobe-2.0
%%Creator: SynPlot.pm
%%Title: $name
%%CreationDate: $date
%%Pages: 1
%%BoundingBox: $b_b $b_b $b_x $b_y
%%Orientation: Portrait
%%PaperSize: A4
%%DocumentNeededFonts: Helvetica-Bold
%%DocumentData: Clean8Bit
%%LanguageLevel: 1
%%EndComments
%%BeginProlog

%CenterShow -- remember that the "stringwidth" operator
% returns two numbers (which allows postscript to deal with
% fonts, such as chinese, which go down the page).
/CS {
    dup stringwidth
    2 div neg exch 2 div neg exch
    rmoveto show
} bind def

%LeftShow
/LS {
    dup stringwidth
    neg exch neg exch
    rmoveto show
} bind def

%%EndProlog
%%Page: 1 1

});
    if ($plot->landscape) {
        $plot->output("
$width 0 translate
90 rotate
");
    }
}

sub footer {
    my( $plot ) = @_;
    
    $plot->output(qq(
showpage
%%Trailer
%%EOF

));

}


## Non-object subroutines ##

BEGIN {
    my @months = qw(
        Jan Feb Mar Apr May Jun
        Jul Aug Sep Oct Nov Dec
    );
    
    sub creation_date {
        my($mday,$mon,$year) = (localtime(time))[3,4,5];
        
        $mon = $months[$mon];
        $year += 1900;
        return "$mday $mon $year";
    }
}

1;

__END__

=head1 NAME - SynPlot

=head1 AUTHOR

James Gilbert B<email> jgrg@sanger.ac.uk

