
# SynPlot - visualization of alignments of syntenic
#           genomic dna sequences using PostScript (TM)
# 
# Copyright (C) 2000 James G.R. Gilbert
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

### SynPlot::Feature ###

package SynPlot::Feature;

use strict;
use Carp;
use SynPlot::State;
use vars '@ISA';
@ISA = ('SynPlot::State');

sub draw {
    confess "Features are only drawn by Sequence object calling the draw_features() method";
}

# height() returning 0 causes draw method
# in SynPlot.pm to ignore this row type.
sub height { return 0 };

# Sequence object is needed for drawing
sub sequence {
    my( $seq, $feat ) = @_;
    
    if ($feat) {
        confess "argument to sequence() must be a 'SynPlot::Sequence'"
            unless $feat->isa('SynPlot::Sequence');
        $seq->{'_sequence'} = $feat;
    }
    return $seq->{'_sequence'};
}

sub name {
    my( $feat, $name ) = @_;
    
    if ($name) {
        $feat->{'_name'} = $name;
    }
    return $feat->{'_name'} || confess "Name not set";
}

BEGIN {

    #  source
    #  feature
    #  start
    #  end
    #  score
    #  strand
    #  frame
    #  group

    my @fields = qw( source
                     feature
                     start
                     end
                     score
                     strand
                     frame
                     group
                     );

    my @field_names = map "_$_", @fields;

    sub new {
        my( $pkg ) = @_;

        my $feat = bless {}, $pkg;
        foreach my $arry (@field_names) {
            $feat->{$arry} = [];
        }
        return $feat;
    }

    # Count returns the number of features in the object
    {
        my $tot_array = $field_names[0];
        sub count {
            my( $feat ) = @_;
    
            return scalar @{$feat->{$tot_array}};
        }

        sub next_feature {
            my( $feat ) = @_;

            my $i = ++$feat->{'_pointer'};
            if (defined $feat->{$tot_array}[$i]) {
                return 1;
            } else {
                $feat->{'_pointer'} = -1;
                return;
            }
        }
    }

    # Generate field access functions
    for (my $i = 0; $i < @fields; $i++) {
        my $get_func   = $fields[$i];
        my $set_func   = "set_$get_func";
        my $field_name = $field_names[$i];
        
        {
            no strict 'refs';
            *$get_func = sub {
                my( $feat ) = @_;
                
                return $feat->{$field_name}[$feat->{'_pointer'}];
            };
            
            *$set_func = sub {
                my( $feat, $value ) = @_;
                
                return $feat->{$field_name}[$feat->{'_pointer'}] = $value;
            };
        }
    }
    
    sub start_array {
        my( $feat ) = @_;
        
        return $feat->{'_start'};
    }
    
    sub delete_all_features {
        my( $feat, @line ) = @_;
        
        for (my $i = 0; $i < @field_names; $i++) {
            $feat->{$field_names[$i]} = [];
        }
    }
    
    sub add_line {
        my( $feat, @line ) = @_;
        
        for (my $i = 0; $i < @field_names; $i++) {
            push @{$feat->{$field_names[$i]}}, $line[$i];
        }
        $feat->{'_pointer'} = $feat->count - 1;
    }
    
    sub write_gff {
        my( $feat, $fh ) = @_;
        
        my $name = $feat->name;
        my $count = $feat->count;
        print $fh "##gff-version 1\n";
        for (my $i = 0; $i < $count; $i++) {
            print $fh join("\t", ($name, map($feat->{$_}[$i], @field_names))), "\n";
        }
    }

    sub read_gff {
        my( $feat, $fh ) = @_;
        
        my $name = $feat->name;
        while (<$fh>) {
            next if /(^\s*#|^$)/;
            chomp;
            my($n, @line) = split(/\s+/, $_, 9);
            next unless $n eq $name;
            $feat->add_line(@line);
        }
    }
    
    sub delete_feature {
        my( $feat ) = @_;
        
        foreach my $f (@field_names) {
            splice(@{$feat->{$f}}, $feat->{'_pointer'}, 1);
        }
    }

    sub zero_pointer {
        my( $feat ) = @_;
        
        $feat->{'_pointer'} = -1;
    }

    sub pointer {
        my( $feat ) = @_;

        my $i = $feat->{'_pointer'};
        if ($i == -1) {
            confess "Pointer not set";
        } else {
            return $i;
        }
    }

    sub set_pointer {
        my( $feat, $i ) = @_;

        my $count = $feat->count;
        if ($i < $count) {
            $feat->{'_pointer'} = $i;
            return 1;
        } else {
            confess "Pointer set off end of features";
        }
    }
}

# Generate config functions
BEGIN {
    my @types = qw( source feature );

    # Color functions
    foreach my $type (@types) {
        no strict 'refs';
        my $func = "${type}_color";
        my $set_func = "set_$func";
        my $type_field = "_${type}_color";

        *$func = sub {
            my( $feat ) = @_;
            
            my $name = $feat->$type();
            return $feat->{'_draw_config'}{$type_field}{$name};
        };

        *$set_func = sub {
            my( $feat, $name, $v ) = @_;
            
            $v = $feat->unix_color_to_ps_rgb($v);
            $feat->{'_draw_config'}{$type_field}{$name} = $v;
        };
    }
    sub color {
        my( $feat ) = @_;
        
        return $feat->source_color || $feat->feature_color || '0 0 0';
    }
    
    # Box height functions
    foreach my $type (@types) {
        no strict 'refs';
        my $func = "${type}_box_height";
        my $set_func = "set_$func";
        my $type_field = "_${type}_box_height";

        *$func = sub {
            my( $feat ) = @_;
            
            my $name = $feat->$type();
            return $feat->{'_draw_config'}{$type_field}{$name};
        };

        *$set_func = sub {
            my( $feat, $name, $v ) = @_;
            
            $v = $v / 100;
            $feat->{'_draw_config'}{$type_field}{$name} = $v;
        };
    }
    sub box_height {
        my( $feat ) = @_;
        
        my $h = $feat->sequence->height;
        my $fraction =  $feat->source_box_height 
            || $feat->feature_box_height
            || 0.75;
        
        my $box_height = $h * $fraction;
        my $y_start  = sprintf("%4.3f", (($h - $box_height) / 2));
        my $y_finish = sprintf("%4.3f", $y_start + $box_height);
                
        return($y_start, $y_finish);
    }
    
    # Layer functions
    foreach my $type (@types) {
        no strict 'refs';
        my $func = "${type}_layer";
        my $set_func = "set_$func";
        my $type_field = "_${type}_layer";
        
        *$func = sub {
            my( $feat ) = @_;

            my $name = $feat->$type();
            return $feat->{'_draw_config'}{$type_field}{$name} || 0;
        };

        *$set_func = sub {
            my( $feat, $name, $v ) = @_;

            confess "Not a plain int '$v'" unless $v =~ /^\d+$/;

            $feat->{'_draw_config'}{$type_field}{$name} = $v;
        };
    }
    sub layer {
        my( $feat ) = @_;
        
        return( $feat->feature_layer, $feat->source_layer );
    }
}

sub store_draw_config {
    my( $feat, $conf ) = @_;
    
    foreach my $type (keys %$conf) {
        my $type_v = $conf->{$type};
        foreach my $name (keys %$type_v) {
            my $name_v = $type_v->{$name};
            foreach my $attribute (keys %$name_v) {
                my $value = $name_v->{$attribute};
                
                my $func = "set_${type}_$attribute";
                $feat->$func($name, $value);
            }
        }
    }
}

sub read_draw_config_file {
    my( $feat, $file ) = @_;
    
    my $conf = do $file;
    $feat->store_draw_config($conf);
}

sub next_active_feature {
    my( $feat ) = @_;

    my $v = $feat->{'_active'}[ ++$feat->{'_active_pointer'} ];
    if (defined($v)) {
        $feat->set_pointer($v);
        return 1;
    } else {
        $feat->zero_pointer;
        $feat->{'_active_pointer'} = -1;
        return;
    }
}

sub set_active {
    my( $feat, @active ) = @_;
        
    my $count = $feat->count;
    $feat->{'_active'} = [];
    $feat->{'_active_pointer'} = -1;
    foreach my $a (@active) {
        if ($a < $count) {
            push @{$feat->{'_active'}}, $a;
        } else {
            confess("Index '$a' is off end feature list of length '$count'");
        }
    }
}

sub make_active_set {
    my( $feat, $x, $y ) = @_;
        
    $feat->zero_pointer;
    
    my( @active_indices );
    while ($feat->next_feature) {
        last if $feat->start > $y;
        if ($feat->overlaps_coords($x, $y)) {
            # $feat->layer returns a list of two values
            push(@active_indices, [$feat->pointer, $feat->layer]);
        }
    }
    # Sort, so that features are drawn in the correct order.
    @active_indices = map $_->[0], 
        sort {$a->[1] <=> $b->[1]
            || $a->[2] <=> $b->[2]} @active_indices;
    
    $feat->set_active(@active_indices);
}

# Strategy:
#
# * Sort features by start then end (sort_gff), and
#   get the index of the first feature who's left
#   hand end is outside the active zone (end_i).
#
# * Sort indices from above by end then start
#   (sort_gff_index), storing lists of indices (r_index),
#   and ends.
#
# * Walk along the r_index array, testing for overlaps
#   of each feature.  Stop when we get to stop_i.

sub debug_show_posn {
    my( $feat, $i ) = @_;
    
    $feat->set_pointer($i);
    my $start = $feat->start;
    my $end   = $feat->end;
    return sprintf "At: %10d  %10d -> %10d\n", $i, $start, $end;
}

sub new_make_active_set {
    my( $feat, $x, $y ) = @_;
    
    #print STDERR "\nFinding features between $x and $y\n";
    
    # Not needed?
    $feat->zero_pointer;
    
    my $end_idx = $feat->_end_index_array
        or confess "No end-index array : Features not sorted?";
    my $end_posn = $feat->_end_posn_array
        or confess "No end-position array : Features not sorted?";
    
    # We stop when we get to the first feature who's
    # start falls after the end of the active zone.
    my $end_i = $feat->index_next_highest($y, $feat->start_array);
    #{
    #    $feat->set_pointer($end_i);
    #    my $count = $feat->count;
    #    my $e_start = $feat->start;
    #    for (my $i = $end_i; $i < $count; $i++) {
    #        print STDERR $feat->debug_show_posn($i);
    #        last if $feat->start != $e_start;
    #    }
    #}
    
    my( @active_indices );
    # Look through the features indexed by their ends,
    # starting at the feature who's end is before the
    # start of the active zone.
    for (my $i = $feat->index_next_lowest($x, $end_posn); $i < @$end_idx; $i++) {
        #my $point = $end_idx->[$i];
        #last if $point == $end_i;
        last if $i == $end_i;
        $feat->set_pointer($i);
        #if ($feat->_debug_overlaps_coords($x, $y)) {
        if ($feat->overlaps_coords($x, $y)) {
            # $feat->layer returns a list of two values
            push(@active_indices, [$i, $feat->layer]);
        }
    }

    # Sort, so that features are drawn in the correct order.
    @active_indices = map $_->[0], 
        sort {$a->[1] <=> $b->[1]
            || $a->[2] <=> $b->[2]} @active_indices;
    #print STDERR "Active = @active_indices\n";
    $feat->set_active(@active_indices);
}

sub overlaps_coords {
    my( $feat, $x, $y ) = @_;
    
    my @pos = ($feat->start, $feat->end);
    
    foreach my $p (@pos) {
        return 1 if $p >= $x and $p <= $y;
    }
    return 1 if $pos[0] < $x and $pos[1] > $y;
    
    return;
}

sub _debug_overlaps_coords {
    my( $feat, $x, $y ) = @_;
    
    my @pos = ($feat->start, $feat->end);
    
    print STDERR "Testing ($x,$y) wether @pos overlaps: ";
    
    my $o = 0;
    foreach my $p (@pos) {
        $o = 1 if $p >= $x and $p <= $y;
    }
    unless ($o) {
        $o = 1 if $pos[0] < $x and $pos[1] > $y;
    }
    
    print STDERR "$o\n";
    
    return $o;
}

sub read_gff_file {
    my( $feat, $file ) = @_;
    
    local *GFF;
    open GFF, $file or confess "Can't read GFF file '$file' : $!";
    $feat->read_gff(\*GFF);
    close GFF;
}

sub write_gff_file {
    my( $feat, $file ) = @_;
    
    local *GFF;
    open GFF, "> $file" or confess "Can't write to GFF file '$file' : $!";
    $feat->write_gff(\*GFF);
    close GFF;
}

sub filter_gff_file {
    my( $feat, $file ) = @_;
    
    local *GFF;
    open GFF, $file or confess "Can't read GFF file '$file' : $!";
    $feat->filter_gff(\*GFF);
    close GFF;
}

    
sub read_gff_index {
    my( $feat, $fh ) = @_;

    my $count = $feat->count;
    my(@idx, @end);
    while (<$fh>) {
        my($i, $start, $end) = split;
        push(@idx, $i);
        push(@end, $end);
    }
    $feat->_end_index_array(\@idx);
    $feat->_end_posn_array (\@end);
}

sub _end_index_array {
    my( $feat, $array ) = @_;
    
    if ($array) {
        confess "Not a ref to an array '$array'"
            unless ref($array) eq 'ARRAY';
        my $count = $feat->count;
        my $array_length = @$array;
        confess "Have '$count' features, but array supplied has '$array_length' elements"
            unless $count == $array_length;
        $feat->{'_end_index'}{'_index'} = $array;
    }
    return $feat->{'_end_index'}{'_index'};
}

sub _end_posn_array {
    my( $feat, $array ) = @_;
    
    if ($array) {
        confess "Not a ref to an array '$array'"
            unless ref($array) eq 'ARRAY';
        my $count = $feat->count;
        my $array_length = @$array;
        confess "Have '$count' features, but array supplied has '$array_length' elements"
            unless $count == $array_length;
        $feat->{'_end_index'}{'_posn'} = $array;
    }
    return $feat->{'_end_index'}{'_posn'}
}

sub write_gff_index {
    my( $feat, $fh ) = @_;

    my $count = $feat->count;
    for (my $i = 0; $i < $count; $i++) {
        print $fh join("\t", ($i, map($feat->{$_}[$i], qw{ _start _end }))), "\n";
    }
}

sub read_gff_index_file {
    my( $feat, $file ) = @_;
    
    local *IDX;
    open IDX, $file or confess "Can't read IDX file '$file' : $!";
    $feat->read_gff_index(\*IDX);
    close IDX;
}

sub write_gff_index_file {
    my( $feat, $file ) = @_;
    
    local *IDX;
    open IDX, "> $file" or confess "Can't write to IDX file '$file' : $!";
    $feat->write_gff_index(\*IDX);
    close IDX;
}

sub sort_gff {
    my( $feat ) = @_;

    my $tmp_dir = "Synplot_Feature.$$.tmp";
    mkdir($tmp_dir, 0777) or confess "Can't mkdir('$tmp_dir') : $!";

    my $before = "$tmp_dir/BEFORE";
    my $after  = "$tmp_dir/AFTER";

    $feat->write_gff_file($before);
    $feat->delete_all_features;
    my $sort_command = "sort -k4n -k5n $before > $after";
    system($sort_command) == 0 or die "sort '$sort_command' failed : $!";
    $feat->read_gff_file($after);
    
    unlink($before, $after);
    rmdir($tmp_dir) or confess "Can't rmdir('$tmp_dir') : $!";
}


sub new_sort_gff {
    my( $feat ) = @_;

    my $tmp_dir = "Synplot_Feature.$$.tmp";
    mkdir($tmp_dir, 0777) or confess "Can't mkdir('$tmp_dir') : $!";

    my $before = "$tmp_dir/BEFORE";
    my $after  = "$tmp_dir/AFTER";

    # Sort features by left then right ends
    {
        $feat->write_gff_file($before);
        $feat->delete_all_features;
        my $sort_command = "sort -k4n -k5n $before > $after";
        system($sort_command) == 0 or die "sort '$sort_command' failed : $!";
        $feat->read_gff_file($after);
    }
    unlink($before, $after) == 2
        or confess "Can't remove '$before' and '$after' : $!";
    
    # Make index of features sorted by right, then left hand ends
    {
        $feat->write_gff_index_file($before);
        my $sort_command = "sort -k3n -k2n $before > $after";
        system($sort_command) == 0 or die "sort '$sort_command' failed : $!";
        $feat->read_gff_index_file($after);
    }
    
    unlink($before, $after);
    rmdir($tmp_dir) or confess "Can't rmdir('$tmp_dir') : $!";
}



foreach my $func (qw( gff_filter )) {
    my $field = "_$func";
    {
        no strict 'refs';
        
        *$func = sub {
            my( $obj, $arg ) = @_;
            
            if (defined($arg)) {
                $obj->{$field} = $arg;
            }
            return $obj->{$field};
        }
    }
}

sub filter_gff {
    my( $feat, $fh ) = @_;
    
    my $filter = $feat->gff_filter || \&default_gff_filter;
    
    my $name = $feat->name;
    while (<$fh>) {
        next if /(\s*#|^$)/;
        chomp;
        my($n, @line) = split(/\s+/, $_, 9);
        next unless $n eq $name;
        $feat->add_line(@line);
        # New feature gets deleted unless &$filter returns TRUE
        $feat->delete_feature unless &$filter($feat);
    }
}

# Gets exons from supported genes and repeats from
# the standard acedb gff dump
sub default_gff_filter {
    my( $feat ) = @_;

    my $source = $feat->source;
    if ($source =~ /supported_/ and $feat->feature eq 'exon') {
        return 1;
    }
    elsif ($source =~ /RepeatMasker/ or $source eq 'tandem' or $source eq 'inverted') {
        $feat->set_feature('repeat');
        return 1;
    }
    else {
        return;
    }
}

sub draw_features {
    my( $feat ) = @_;
    
    my $seq = $feat->sequence or confess "Can't draw features without sequence";
        
    my $seq_start = $seq->seq_coord($seq->a);
    my $seq_end   = $seq->seq_coord($seq->b);
    
    my $from = $feat->from;
    
    my $active_left  = $feat->coord($seq->a - $from);
    my $active_right = $feat->coord($seq->b - $from);
    
    # Hack that was put in for GLASS alignment
    # -- solved in glass_to_fasta script
    #my $active_length = $seq_end - $seq_start + 1;
    #if ($active_length < 50) {
    #    warn "Skipping active length of $active_length\n";
    #    return;
    #}
    
    # make_active_set() finds which features overlap the
    # current active region, and arranges them in the
    # order in which they should be drawn, from the bottom
    # layer to the top layer.
    $feat->make_active_set($seq_start, $seq_end);
    
    my $active_color = '';
    $feat->output("gsave\n");
    while ($feat->next_active_feature) {
        
        #warn join(" ", map {"$_=". $feat->$_()} qw(source start end group)), "\n";
        
        #if ($feat->group eq 'hsscl.2.mRNA') {
        #    warn "Drawing hsscl exon with sequence ",
        #        join(" ", map {"$_=". $feat->$_()} qw(start end)), "\n";
        #}
        
        # Get the color for this feature, and set it
        # to the active color if it isn't already set
        my $c = $feat->color;
        #unless ($c eq $active_color) {
        #    $feat->output("$c setrgbcolor\n");
        #    $active_color = $c;
        #}
        $feat->output("$c setrgbcolor\n");
        
        # Calculate the height of the feature
        my($y_start, $y_finish) = $feat->box_height;
        
        # Calculate the left and right edges of the feature
        my $A = $feat->coord( $seq->aln_coord($feat->start) - $from );
        my $B = $feat->coord( $seq->aln_coord($feat->end)   - $from );
        $A = $active_left  if $A < $active_left;
        $B = $active_right if $B > $active_right;
        $seq->output(qq{
newpath
$A $y_start moveto
$A $y_finish lineto
$B $y_finish lineto
$B $y_start lineto
closepath
fill
});
    }
    $feat->output("grestore\n");
}

1;

__END__

=head1 NAME - SynPlot::Feature

=head1 AUTHOR

James Gilbert B<email> jgrg@sanger.ac.uk

