
# SynPlot - visualization of alignments of syntenic
#           genomic dna sequences using PostScript (TM)
# 
# Copyright (C) 2000 James G.R. Gilbert
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

### SynPlot::Sequence ###

package SynPlot::Sequence;

use strict;
use Carp;
use SynPlot::State;
use vars '@ISA';
@ISA = ('SynPlot::State');

sub new {
    my( $pkg ) = @_;
    
    return bless {}, $pkg;
}

foreach my $func (qw( name dna )) {
    my $field = "_$func";
    {
        no strict 'refs';
        
        *$func = sub {
            my( $obj, $arg ) = @_;
            
            if (defined($arg)) {
                $obj->{$field} = $arg;
            }
            return $obj->{$field};
        }
    }
}

sub seq_length {
    my( $seq ) = @_;

    return length($seq->dna);
}

sub height { return 12 }

# So I don't have to remember which array is which
# in the gap map.
BEGIN {
    sub GAP () { 0 }
    sub ALN () { 1 }
    sub SEQ () { 2 }
}

sub gap_map {
    my( $seq ) = @_;

    unless ($seq->{'_gap_map'}) {
        my $dna = $seq->dna;
        
        my( @gap,   # Cumulative counts of gap lengths
            @aln,   # Positions of gap end in alignment
            @seq,   # Last positions in sequence before alignment
            );
        for (my $i = 0; $dna =~ /-+/g; $i++) {
            my $pre = length($`);
            my $mat = length($&);
            
            $aln[$i] = $pre + $mat;
            if ($i == 0) {
                $gap[$i] = $mat;
                $seq[$i] = $pre;
            } else {
                $gap[$i] = $mat + $gap[$i-1];
                $seq[$i] = $aln[$i] - $gap[$i];
            }
        }

        my $map = [];
        $map->[GAP] = \@gap;
        $map->[ALN] = \@aln;
        $map->[SEQ] = \@seq;
        $seq->{'_gap_map'} = $map;
    }

    return $seq->{'_gap_map'};
}

# Returns the coordinate in the aligned sequence for
# the supplied genomic coordinate
sub aln_coord {
    my( $seq, $x ) = @_;
    
    my $m = $seq->gap_map;
    my $i = $seq->index_next_lowest($x, $m->[SEQ]);
    return $i == -1 ? $x : $x + $m->[GAP][$i];
}

# Returns the coordinate in the genome sequence for
# the supplied alignment coordinate
sub seq_coord {
    my( $seq, $x ) = @_;
    
    if ($seq->is_gap($x)) {
        confess("Can't give position of the gap at '$x' in genomic sequence!\n");
    }
    
    my $m = $seq->gap_map;
    my $i = $seq->index_next_lowest($x, $m->[ALN]);
    return $i == -1 ? $x : $x - $m->[GAP][$i];
}

# Returns true if the position in the sequence ($x) is a gap.
sub is_gap {
    my( $seq, $x ) = @_;
    
    if (substr($seq->dna, $x - 1, 1) eq '-') {
        return 1;
    } else {
        return 0;
    }
}

# This block contains methods for iterating accross the sequence
# regions in a row
BEGIN {
    my $A = '_active_start';
    my $B = '_active_end';
    my $I = '_active_cursor';

    # Stores a map of segments to be drawn on the current row
    sub set_active {
        my( $seq ) = @_;

        my $m = $seq->gap_map;
        my $from = $seq->from;
        my $upto = $seq->upto;
                
        # Make new arrays for storing segements,
        # and set pointer to current
        my( @start, @end );
        $seq->{$A} = \@start;
        $seq->{$B} = \@end;
        $seq->{$I} = -1;
        
        # $i gets -1 if there are no gaps in front of this position.
        my $i = $seq->index_next_highest($from, $m->[ALN]);

        # Are we starting in a gap
        if ($seq->is_gap($from)) {
            my $e = $seq->gap_end($i);
            
            # Row may be entirely gap
            if ($e > $upto) {
                return 1;
            }
            
            push(@start, $seq->gap_end($i));
        } else {
            push(@start, $from);
        }
                
        # Loop through remaining gaps on this row
        while (defined( $i = $seq->index_next_highest($start[$#start], $m->[ALN]) )) {
            last if $i == -1;
            
            my $s = $seq->gap_start($i);
            last if $s > $upto;
            push(@end, $s);
            
            my $e = $seq->gap_end($i);
            
            # Finished if the next gap is off the end
            return 1 if $e > $upto;
            
            push(@start, $e);
        }
        
        # End must be end of row
        push(@end, $upto);
        return 1;
    }
    
    sub next_active {
        my( $seq ) = @_;
        
        my $i = ++$seq->{$I};   # Move pointer to next element
        if (defined($seq->{$A}[$i])) {
            $seq->{$I} = $i;
            return 1;
        } else {
            $seq->{$I} = -1;
            return;
        }
    }

    sub a {
        my( $seq ) = @_;
        return $seq->{$A}[$seq->{$I}];
    }
    sub b {
        my( $seq ) = @_;
        return $seq->{$B}[$seq->{$I}];
    }
    
    # For debuggin
    sub show_active_array {
        my( $seq ) = @_;
        
        my $str;
        foreach ($A, $B) {
            my $pat = join(' ', map '%4d', @{$seq->{$_}});
            $str .= sprintf "%15s $pat\n", $_, @{$seq->{$_}};
        }
        return $str;
    }
}

# Returns the length of the gap at $i in the map
sub gap_length {
    my( $seq, $i ) = @_;
    
    my $m = $seq->gap_map;
    if ($i == 0) {
        # There's no previous element if we're at
        # the start of the gap map
        return $m->[GAP][$i]
    } else {
        return $m->[GAP][$i] - $m->[GAP][$i - 1];
    }
}

# Returns the position of the last base before the
# start of the gap
sub gap_start {
    my( $seq, $i ) = @_;
    
    my $m = $seq->gap_map;
    return $m->[ALN][$i] - $seq->gap_length($i);
}

# Returns the position of first base after the gap
sub gap_end {
    my( $seq, $i ) = @_;
    
    my $m = $seq->gap_map;
    return $m->[ALN][$i] + 1;
}

sub draw {
    my( $seq, $plot ) = @_;
    
    my $feat = $plot->get_Feature($seq->name);
    # Make sure the feature object knows about this sequence
    $feat->sequence($seq) if $feat;

    my $centre = $seq->height / 2;
    my $scale = $seq->scale;
    my $from = $seq->from;
    $seq->set_active;
    
    $seq->output("
% Start of sequence line

1 setlinewidth
0 setlinecap
0 setgray
");
    
    while ($seq->next_active) {
        my $x = $seq->coord($seq->a - $from);
        my $y = $seq->coord($seq->b - $from);
        $seq->output("
newpath
$x $centre moveto
$y $centre lineto
stroke
");
        # Draw sequence features
        $feat->draw_features if $feat;
    }

    $seq->output("
% End of sequence line
");

}

1;

__END__

=head1 NAME - SynPlot::Sequence

=head1 AUTHOR

James Gilbert B<email> jgrg@sanger.ac.uk

