
# SynPlot - visualization of alignments of syntenic
#           genomic dna sequences using PostScript (TM)
# 
# Copyright (C) 2000 James G.R. Gilbert
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

### SynPlot::Summary ###

package SynPlot::State;

use strict;
use Carp;

sub new {
    my( $pkg ) = @_;
    
    return bless {}, $pkg;
}

sub height {
    my( $ele ) = @_;
    
    my $ref = ref($ele);
    warn "Using dummy height for $ref\n";
    return 12;
}

sub draw {
    my( $ele ) = @_;
    
    my $ref = ref($ele);
    warn "Using dummy draw for $ref\n";
    
    my $len = $ele->point_length;
    $ele->output(qq(
% Dummy row box
newpath
0 0 moveto
0 10 lineto
$len 10 lineto
$len 0 lineto
closepath
.8 setgray
fill
));
}

sub point_length {
    my( $ele ) = @_;
    
    return sprintf "%1.1f", $ele->row_length * $ele->scale;
}

sub round {
    my( $ele, $x ) = @_;
    
    return sprintf "%3.2f", $x;
}

sub coord {
    my( $ele, $x ) = @_;
    
    return sprintf "%3.2f", $x * $ele->scale;
}

sub state {
    my( $ele, $arg ) = @_;

    if ($arg) {
        confess "Not a SynPlot::State object"
            unless $arg->isa('SynPlot::State');
        $ele->{'_state'} = $arg;
    }
    return $ele->{'_state'};
}

sub magnification {
    my( $ele, $arg ) = @_;

    if ($arg) {
        confess('Only an SynPlot object is permitted to set magnification')
            unless $ele->isa('SynPlot');
        $ele->state->{'_magnification'} = $arg;
    }
    return $ele->state->{'_magnification'} || 1;
}

# Performs a binary search through @$a, returning the index
# of the number closest to, but smaller than, $n.
# Returns -1 if none such
sub index_next_lowest {
    my( $ele, $n, $a ) = @_;
    
    my $x = 0;
    my $y = @{$a};
    
    my( $j );
    while ($x < $y) {
        use integer;
        my $j = ($y + $x) / 2;
        
        if ($a->[$j] < $n) {
            $x = $j + 1;
        }
        else {
            $y = $j;
        }
    }
    return $x - 1;
}

# As above, but returns index of number equal to $n, or
# the next largest.
# Returns -1 if none such
sub index_next_highest {
    my( $ele, $n, $a ) = @_;
    
    my $x = 0;
    my $y = @{$a};
    
    my( $j );
    while ($x < $y) {
        use integer;
        my $j = ($y + $x) / 2;
        
        if ($a->[$j] < $n) {
            $x = $j + 1;
        }
        else {
            $y = $j;
        }
    }
    return $x == @{$a} ? -1 : $x;
}

BEGIN {

    foreach my $func (qw( from
                          upto
                          scale
                          nuc_width
                          row_number
                          row_length
                          page_number
                          canvas_width
                          canvas_height
                          )) {
        my $field = "_$func";
        {
            no strict 'refs';

            *$func = sub {
                my( $obj, $arg ) = @_;

                if (defined($arg)) {
                    confess("Only an SynPlot object is permitted to set $func")
                        unless $obj->isa('SynPlot');
                    $obj->state->{$field} = $arg;
                }
                return $obj->state->{$field};
            }
        }
    }

    foreach my $func (qw( left_offset
                          right_offset
                          )) {
        my $field = "_$func";
        {
            no strict 'refs';

            *$func = sub {
                my( $obj, $arg ) = @_;

                if (defined($arg)) {
                    confess("Only an SynPlot object is permitted to set $func")
                        unless $obj->isa('SynPlot');
                    confess("Not a positive integer '$arg'")
                        unless $arg =~ /^\d+$/;
                    $obj->state->{$field} = $arg;
                }
                return $obj->state->{$field};
            }
        }
    }
}

sub offsets {
    my( $ele ) = @_;

    my $left  = $ele->left_offset;
    my $right = $ele->right_offset;

    #confess "Offsets not set: left='$left' right='$right'"
    return unless $left and $right;
    if ($left < $right) {
        return($left,$right);
    } else {
        confess "Bad offsets: left='$left' is not less than right='$right'";
    }
}

sub offset_length {
    my( $ele ) = @_;
    
    my($l,$r) = $ele->offsets or return;
    return $r - $l + 1;
}

# These functions probably shouldn't be in here
{
    my $field = '_output';

    sub set_output {
        my( $obj, $arg ) = @_;

        confess('Only an SynPlot object is permitted to call set_output')
            unless $obj->isa('SynPlot');
        if ($arg) {
            my $ref = ref($arg);
            if ($ref eq 'SCALAR') {
                $obj->state->{$field} = $arg;
            }
            elsif ($ref eq 'GLOB') {
                $obj->state->{$field} = $arg;
            }
            else {
                confess("Argument to set_output is a '$ref' ref; must be 'SCALAR' or 'GLOB'");
            }
        }
        return $obj->state->{$field};
    }

    sub output {
        my $obj = shift;

        my $out = $obj->state->{$field}
            or confess("You haven't defined the output with set_output");
        if (ref($out) eq 'SCALAR') {
            $$out .= join('', @_);
        } else {
            print $out @_;
        }
    }
}


{
    my( %color );
    
    sub unix_color_to_ps_rgb {
        my( $ele, $unix_color_name ) = @_;
        $unix_color_name = lc $unix_color_name;
        
        unless (%color) {
            local *RGBCOLOR;
            open RGBCOLOR, "showrgb |"
                or confess "Can't open pipe from 'showrgb' : $!";
            while (<RGBCOLOR>) {
                my @fields = split;
                my @rgb = @fields[0,1,2];
                my $name = join(' ', @fields[3..$#fields]);
                foreach my $v (@rgb) {
                    $v = sprintf("%4.3f", $v / 255);
                }
                $color{lc $name} = join(' ', @rgb);
            }
        }
        return $color{$unix_color_name} || confess "No such color '$unix_color_name'";
    }
}


1;

__END__

=head1 NAME - SynPlot::Summary

=head1 AUTHOR

James Gilbert B<email> jgrg@sanger.ac.uk

