
# SynPlot - visualization of alignments of syntenic
#           genomic dna sequences using PostScript (TM)
# 
# Copyright (C) 2000 James G.R. Gilbert
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

### SynPlot::Summary ###

package SynPlot::Summary;

use strict;
use Carp;
use SynPlot::State;
use vars '@ISA';
@ISA = ('SynPlot::State');

sub new {
    my( $pkg ) = @_;
    
    return bless {}, $pkg;
}

sub height {
    my( $summ, $arg ) = @_;
    
    if ($arg) {
        $summ->{'_height'} = $arg;
    }
    return $summ->{'_height'} || 60;
}

sub sequence_count {
    my( $summ, $count ) = @_;
    
    if ($count) {
        confess "Not an integer '$count'"
            unless $count =~ /^\d+$/;
        $summ->{'_sequence_count'} = $count;
    } else {
        $count = $summ->{'_sequence_count'};
    }
    
    if ($count) {
        return $count;
    } else {
        confess "No sequence count stored";
    }
}

sub window_size {
    my( $summ, $arg ) = @_;
    
    if ($arg) {
        $summ->{'_window_size'} = $arg;
    }
    return $summ->{'_window_size'} || 100;
}

sub increment {
    my( $summ, $arg ) = @_;
    
    if ($arg) {
        $summ->{'_increment'} = $arg;
    }
    return $summ->{'_increment'} || 50;
}

sub bitmap {
    my( $summ, $arg ) = @_;
    
    if ($arg) {
        $summ->{'_bitmap'} = $arg;
    }
    return $summ->{'_bitmap'};
}

sub analyse {
    my( $summ, @sequence ) = @_;
    
    $summ->sequence_count(scalar @sequence);
    confess "Got less than two sequences" unless @sequence > 1;
    
    # Get and check sequence lengths
    my( $length );
    foreach my $s (@sequence) {
        my $l = $s->seq_length;
        $length ||= $l;
        confess "Sequence length mismatch: $l vs $length"
            unless $l == $length;
    }
    
    my @dna = map $_->dna, @sequence;
    
    my $map = 0 x $length;
    
    # At each position...
    BASE: for (my $i = 0; $i < $length; $i++) {
        my( %c_count );
        # on each sequence...
        for (my $j = 0; $j < @dna; $j++) {
            # get the character at $i
            my $c = substr($dna[$j], $i, 1);
            # and count it (unless it is a gap)
            $c_count{$c}++ unless $c eq '-';
        }
        
        # Record the count of the most common
        # character at this position, minus 1
        my $max = 1;
        foreach my $v (values %c_count) {
            $max = $v if $v > $max;
        }
        $max--;
        substr($map, $i, 1) = $max;
    }
    $summ->bitmap($map);
}

sub old_analyse {
    my( $summ, @sequence ) = @_;
    
    $summ->sequence_count(scalar @sequence);
    confess "Got less than two sequences" unless @sequence > 1;
    
    # Get and check sequence lengths
    my( $length );
    foreach my $s (@sequence) {
        my $l = $s->seq_length;
        $length ||= $l;
        confess "Sequence length mismatch: $l vs $length"
            unless $l == $length;
    }
    
    my @dna = map $_->dna, @sequence;
    
    my $map = 0 x $length;
    # At each position...
    BASE: for (my $i = 0; $i < $length; $i++) {
        my( $p );
        # on each sequence...
        for (my $j = 0; $j < @dna; $j++) {
            my $c = substr($dna[$j], $i, 1);
            next BASE if $c eq '-' or $c eq 'n';
            if ($p) {
                next BASE unless $c eq $p;
            } else {
                $p = $c;
            }
        }
        # set the corresponding position on $map to 1
        # if all bases are the same, and aren't gaps.
        substr($map, $i, 1) = 1;
    }
    $summ->bitmap($map);
}

sub draw {
    my( $summ ) = @_;

    $summ->output("% Start of profile plot\n");
    $summ->draw_thresholdbox;
    $summ->draw_tickmarks;
    $summ->draw_plotline;
    $summ->draw_surroundingbox;
    $summ->output("% End of profile plot\n");

}

sub draw_plotline {
    my( $summ ) = @_;

    my(@pos, @scr);
    {
        my $from = $summ->from;
        my $upto = $summ->upto;
        my $win = $summ->window_size;
        my $win_pattern = 'a' x $win;
        
        # The total score contained within each window
        # is divded by $scale, which is the length of
        # the window multiplied by the number of sequences
        # in the alignmnet, minus 1 (ie: the maximum
        # possible score).
        my $scale = $win * ($summ->sequence_count - 1);

        my $i = $from - 1;          # Middle of window (in
                                    # zero based coordinates).
        my $x = $i - ($win / 2);    # Start of window.
        my $y = $i + ($win / 2);    # End of window.

        # Move forward if $x is off start of $map
        if ($x < 0) {
            my $n = -1 * $x;
            map {$_ += $n} ($i, $x, $y);
        }

        my $inc = $summ->increment;
        my $map = $summ->bitmap;
        my $len = length($map);
        my $off = $from - 1;
        
        # Calculate scores until we hit the end of this
        # row, or the end of the sequence
        while ($i < $upto and $y < $len) {  # Tests are < not <= because
                                            # coordinates are zero based
            
            # Total the numbers in the window
            my $count = 0;
            foreach my $n (unpack $win_pattern, substr($map, $x, $win)) {
                $count += $n;
            }
            my $score = $count / $scale;

            push(@pos, $i - $off);
            push(@scr, $score);

            # Move the pointers forward by the increment.
            map {$_ += $inc} ($i, $x, $y);
        }
    }
    
    my $height = $summ->height;
    my $y_start = $summ->round($scr[0] * $height);
    $summ->output("
% Profile plot line
newpath
0 $y_start moveto
");

    my $penultimate = @pos - 1;
    for (my $i = 1; $i < $penultimate; $i++) {
        my $x = $pos[$i];
        my $y = $scr[$i];
        
        # Skip points which don't change height of plot
        next if $y == $scr[$i-1] and $y == $scr[$i+1];
        
        $x = $summ->coord($x);
        $y = $summ->round($y * $height);
        $summ->output("$x $y lineto\n");
    }

    my $y_end = $summ->round($scr[-1] * $height);
    my $x_end = $summ->coord($pos[-1]);

    $summ->output("
$x_end $y_end lineto
0.25 setlinewidth
0 setgray
1 setlinejoin
stroke

");
    return 1;
}


sub old_draw_plotline {
    my( $summ ) = @_;

    my(@pos, @scr);
    {
        my $from = $summ->from;
        my $upto = $summ->upto;
        my $win = $summ->window_size;

        my $i = $from - 1;          # Middle of window (in
                                    # zero based coordinates).
        my $x = $i - ($win / 2);    # Start of window.
        my $y = $i + ($win / 2);    # End of window.

        # Move forward if $x is off start of $map
        if ($x < 0) {
            my $n = -1 * $x;
            map {$_ += $n} ($i, $x, $y);
        }

        my $inc = $summ->increment;
        my $map = $summ->bitmap;
        my $len = length($map);
        my $off = $from - 1;
        # Calculate scores until we hit the end of this
        # row, or the end of the sequence
        while ($i < $upto and $y < $len) {  # Tests are < not <= because
                                            # coordinates are zero based

            # Count the number of 1's in this bit of sequence
            my $count = substr($map, $x, $win) =~ tr/1/1/;
            my $score = $count / $win;

            push(@pos, $i - $off);
            push(@scr, $score);

            map {$_ += $inc} ($i, $x, $y);
        }
    }
    
    my $height = $summ->height;
    my $y_start = $summ->round($scr[0] * $height);
    $summ->output("
% Profile plot line
newpath
0 $y_start moveto
");

    my $penultimate = @pos - 1;
    for (my $i = 1; $i < $penultimate; $i++) {
        my $x = $pos[$i];
        my $y = $scr[$i];
        
        # Skip points which don't change height of plot
        next if $y == $scr[$i-1] and $y == $scr[$i+1];
        
        $x = $summ->coord($x);
        $y = $summ->round($y * $height);
        $summ->output("$x $y lineto\n");
    }

    my $y_end = $summ->round($scr[-1] * $height);
    my $x_end = $summ->coord($pos[-1]);

    $summ->output("
$x_end $y_end lineto
0.25 setlinewidth
0 setgray
1 setlinejoin
stroke

");
    return 1;
}

sub draw_surroundingbox {
    my( $summ ) = @_;
    
    my $height = $summ->height;
    my $point_length = $summ->point_length;
    $summ->output("
% Box around plot
newpath
0 0 moveto
0 $height lineto
$point_length $height lineto
$point_length 0 lineto
0 0 lineto
closepath
0.25 setlinewidth
0 setgray
stroke
");

}

sub draw_thresholdbox {
    my( $summ ) = @_;
    
    my $seq_count = $summ->sequence_count;
    my $point_length = $summ->point_length; # Row length in points
    my $thresh_height = $summ->height / (4 ** ($seq_count - 1));
    
    
    $summ->output("
% Light grey box behind bottom of plot
newpath
0 0 moveto
0 $thresh_height lineto
$point_length $thresh_height lineto
$point_length 0 lineto
0 0 lineto
closepath
.9 setgray
fill
");
    
}

sub draw_tickmarks {
    my( $summ ) = @_;
    
    # Tick marks under plot
    {
        my $nuc_width = $summ->nuc_width;
        my $major = 1;
        my $max_ticks = 20; # Maximum number of ticks under plot
        while (($nuc_width / $major) > $max_ticks) {
            $major *= 10;
        }
        my $minor = $major / 10;
        my $half  = $major / 2;
        
        my $label_font_size = 6;
        my $tick_pad = 4 + $label_font_size;
        
        $summ->output("
%Tick marks underneath plot
0.25 setlinewidth
0 setgray
/Helvetica-Bold findfont $label_font_size scalefont setfont
");
        
        my $from = $summ->from;
        my $upto = $summ->upto;
        # Find the first base position which gets a tick
        my( $mark );
        if (my $rem = $from % $minor) {
            $mark = $from - $rem + $minor;
        } else {
            $mark = $from;
        }
        for ( ; $mark <= $upto; $mark += $minor) {
            my $x = $summ->coord($mark - $from);
            
            my $mark_text_len = length($mark);
            my $mark_pat = 'a3' x int($mark_text_len / 3);
            if (my $rem = $mark_text_len % 3) {
                $mark_pat = "a$rem$mark_pat";
            }
            my $formatted_mark = join(' ', unpack $mark_pat, $mark);
            
            if (my $rem = $mark % $major) {
                my $tick_len = $rem == $half ? 4 : 2;
                $summ->output("
newpath
$x 0 moveto
$x -$tick_len lineto
stroke
");
            } else {
                $summ->output("
newpath
$x 0 moveto
$x -4 lineto
stroke
$x -$tick_pad moveto
($formatted_mark) CS
");
            }
        }
    }
}

1;

__END__

=head1 NAME - SynPlot::Summary

=head1 AUTHOR

James Gilbert B<email> jgrg@sanger.ac.uk

